import sys

from copy import deepcopy
from time import sleep, monotonic
from collections import defaultdict
from conllu_reader import *
from language_importer import import_data, folder2lang, id2folder
from csv_functions import read_table, list2csv#, tabsep2csv
import Finnish
#import Afrikaans
import Swedish
import Vietnamese
import Chinese
import RomanianHYPHENRRT
#import Hebrew
import RomanianHYPHENNonstandard
import German
import KoreanHYPHENGSD
import KoreanHYPHENKaist
import Russian
import PolishHYPHENLFG
import PolishHYPHENSZ
import Dutch
import Latvian

hard_languages = [w for w in """Finnish
Swedish
Vietnamese
Chinese
Romanian-RRT
#Hebrew
Romanian-Nonstandard
German
Korean-GSD
Korean-Kaist
Russian
Polish-LFG
Polish-SZ
Dutch
Latvian""".splitlines() if not w.startswith('#')] # languages and materials that are not doable with the simple rules allowed by the chart

hard_materials = [w for w in hard_languages if '-' in w]
# cases where there are more than one treebank for a single language
# and possessive NPs are not recognized from them using the same logic


def compare(tag, condition):
    """Determine whether a single tag meets a condition."""
    if condition == '': # empty condition, meaning the content of that particular cell is irrelevant
        return True
    if ';' in condition: # ";" means 'or': if any of the subconditions separated by semicolons are true, the whole condition is true
        return any(compare(tag, c) for c in condition.split(';')) # separate first by "or" conditions and then by "and" conditions
    # That is a recursion. Cannot be infinite as there are no semicolons in a list that was formed by splitting a string at semicolons.

    conditions = condition.split('&')

    answers = []

    for ehto in conditions: # ran out of variable names.
        #"ehto" is Finnish for 'condition'
    
        if ehto.startswith('on.!'): #on.!string = not equal to "string"
            answers.append(tag != ehto[4:])
        elif ehto.startswith('on.'): #on.string = equal to "string"
            answers.append(tag == ehto[3:])
        elif ehto.startswith('sis.!'): #sis.!string = does not contain "string"
            answers.append(ehto[5:] not in tag)
        elif ehto.startswith('sis.'): #sis.!string = contains "string"
            answers.append(ehto[4:] in tag)

    return all(answers) # in order for a tag to meet an "and" condition, all of its subconditions must be true

ehtotaulukko = [] # tag table as list of strings
provisional = set() # languages whose pronouns or nouns we can deal with, but not both. Should be empty by the end of the project.

for line in avaa('tag_chart.txt').strip().splitlines()[1:]:
    if len(line.split()) > 2:
        ehtotaulukko.append(line.split('\t'))
    else: # in case there are two or less non-blank cells*, that means the language is provisional, meaning only pronouns or nouns have been done.
        provisional.add(line.strip().split('\t')[-1]) # When the tag chart is ready, this block should never be reached.
    # *now that the only whitespace characters are tabs. This would likely not work if e.g. there were spaces in the file.

# if a hard language to be provisional, it is indicated there by adding there an otherwise unused variable called "provisional"
for language in hard_languages:
    if 'provisional' in dir(eval(language.replace('-', 'HYPHEN'))):
        provisional.add(language)

print('Provisional languages:', provisional)

#If there are no provisional languages, the following one-line command would create ehtotaulukko but it would not create provisional.
#ehtotaulukko = [line.split('\t') for line in avaa('tag_chart.txt').strip().splitlines()[1:] if len(line.split()) > 2]

condition_ddict = defaultdict(lambda:[])

for line in ehtotaulukko:
    condition_ddict[line[1]].append(line[2:]) # language: conditions
    condition_ddict[line[0]].append(line[2:]) # id: conditions
    condition_ddict[id2folder[line[0]]].append(line[2:]) # folder: conditions
    condition_ddict[id2folder[line[0]][3:]].append(line[2:]) # folder without UD_ : conditions
    # Because of the previous four lines, condition_ddict is really flexible in how you call it.
    

languages = [line[1] for line in ehtotaulukko]
languages = sorted(set(languages), key=lambda j:languages.index(j))

NA = 'NA'

def lemma_comparison(wordform, lemma, language): # language is e.g. Estonian-EDT
    """Compare whether wordform and lemma are the same, taking into account the special markings in the lemma (e.g. "#" for compounds).
Language parameter is the name of the folder without the 'UD_' beginning, for example 'Estonian-EDT'."""
    if language.startswith('Afrikaans'):
        real_wordform = wordform.lower()
        real_lemma = lemma.lower()
        return real_wordform == real_lemma and real_lemma in {'u', 'ons', 'julle', 'hulle'}
    elif language.startswith('Basque'):
        return wordform.lstrip('*').lower() == lemma.lstrip('*').lower()
    elif language.startswith('Estonian'):
        real_wordform = wordform.lower().replace('_', '') # final replacements probably needless here...
        if real_wordform in {'mina', 'ma', 'me', 'meie'} and lemma == 'mina': # I, we
            return True
        if real_wordform in {'sina', 'sa', 'te', 'teie'} and lemma == 'sina': # you
            return True
        if real_wordform in {'nemad', 'nad', 'ta', 'tema'} and lemma == 'tema': # he, she, they
            return True
        return real_wordform == lemma.lower().replace('_', '').replace('=', '') # ... but these replacements [at least the first one] are definitely needed.
    elif language.startswith('Old'):
        real_wordform = wordform.lower()
        lsp = lemma.split('#')
        if len(lsp) > 2:
            print(wordform, lemma, language, sep='\t')
        if lsp[-1].isdigit():
            return real_wordform == lsp[0].lower()
        if len(lsp) != 1:
            print(wordform, lemma, language, 1, sep='\t')
        return real_wordform == lsp[0].replace('#', '').lower()
    elif language.startswith('Ukrainian'):
        rwordform = wordform.lower().replace('́', '').replace("'", '’') # remove stress marks as well. Not sure if last replacement is appropriate,
        # but done because of Запам'ятайте, lemma: запам’ятати in sentence 2uk5
        if lemma.endswith('.'):
            lemma = lemma[:-1]
        return rwordform == lemma.lower()
    elif language.startswith('Italian'):
        return wordform.lower() == lemma == 'loro'
    return wordform.lower() == lemma.lower() # default

def is_done(kieli):
    """Find if there is information on this language's possessive NPs either in the tag table or in a script."""
    if kieli in folder2lang:
        kieli = folder2lang[kieli].replace(' ', '_')

    if kieli in hard_languages:
        try:
            eval(kieli.replace(' ', '_').replace('-', 'HYPHEN') + '.possessive')
            return True
        except NameError:
            return False

    if kieli not in condition_ddict:
        print(kieli, '(not done yet)', end= ' ')
        return False
    
    return True

def adpcase(three_conditions):
    c1, c2, c3 = three_conditions

    if c1 != '':
        return False
    if not (c2.startswith('sis.') or c2.startswith('on.')):
        if not (c3.startswith('sis.') or c3.startswith('on.')):
            return True
    return False

def possessiivi(lause, kieli):
    """Find dependents of possessive NPs n a sentence (lause).
lause is an ordered dict (conllu item), kieli is the language or the material (folder or id).

If the language or material has its own script, it will use that, otherwise it will look for instructions in the tag chart."""

    language_only = kieli.split('-')[0]

    if kieli in hard_materials:
        return eval(kieli.replace('-', 'HYPHEN') + '.possessive(lause, dict, subtract=True)')
    elif language_only in hard_languages:
        return eval(language_only + '.possessive(lause, dict, subtract=True)')

    def find_head_index(word):
        return int(word[6])-1

    answers = []

    conj = 'conj'
    punct = 'PUNCT'

    analyses = sorted(lause['analyysit'], key=lambda i:i[7] == conj)

    if kieli not in condition_ddict:
        return {'dep_marked':NA, 'double_marked':NA, 'head_marked':NA, 'zero_marked':NA, 'head_exist':NA}
    else:
        conditions = condition_ddict[kieli]

        

    answer = defaultdict(lambda:[])


    for ehto in conditions:
            
        if ehto[:3] == ['', '', ''] or adpcase(ehto[:3]):
            if ehto[3:5] != ['', ''] and (ehto[5].startswith('on.') or ehto[5].startswith('sis.') or ehto[5] == ''):
                for dependent in analyses: # "for word in analyses" would be more readable, but this checks if it is a dependent.
                    dep_pos = dependent[3]
                    dep_tag = dependent[5]
                    dep_synt = dependent[7]
                    head_i = find_head_index(dependent)
                    if head_i < 0:
                        continue
                    head = lause['analyysit'][head_i]
                    head_pos = head[3]
                    head_tag = head[5]

                    lst = [dep_pos, dep_tag, dep_synt, head_pos, head_tag]

                    adp = ehto[1] # ADP / case
                    case = ehto[2] # often empty strings, but making possible to filter out adposition constructions
                    # for this to work, both must be non-empty strings

                    if bool(adp) and bool(case):
                        if any([(word[6] == dependent[0] and word[3] == adp and word[7] == case) for word in analyses]):
                            continue
                        elif kieli == 'Old_Church_Slavonic-PROIEL': # disallow noun/proper noun nmods that are before the head
                            if int(dependent[0]) < int(head[0]): # for Old Church Slavonic
                                continue
                            

                    if all(compare(*pair) for pair in zip(lst, ehto[3:])):
                        # Italian: "il loro nido" ('their nest', lit. "the their nest" the head of "il" is "nido" not "loro")
                        # "loro" is both 'they' and 'their' in Italian

                        # For a construction to be zero marked in Italian:
                        # the pronoun must be "loro" as it's the only Italian pronoun that is both a subject and a possessive pronoun
                        # there must be no article (it's lemma seems to be always "il" in all treebanks even with contractions)
                        
                        if '~' in ehto[-1]:
                            plemma, ppos, psynt, instruction = ehto[-1].split('~')
                            assert instruction.startswith('clemma.')
                            class_choices = instruction[7:].split(';')
                            
                            do_lemma_comparison = True

                            for word in analyses: # Check if there is an article in the construction
                                if word[6] != head[0]:
                                    continue
                                if (word[2] == plemma or not plemma) and (word[3] == ppos or not ppos) and (word[7] == psynt or not psynt):
                                    do_lemma_comparison = False
                                    break

                            if do_lemma_comparison:
                                
                                if lemma_comparison(dependent[1], dependent[2], kieli):
                                    answer[class_choices[0] + '_marked'].append(dependent.copy())
                                    #print(class_choices[0] + '_marked', dependent)
                                else:
                                    answer[class_choices[1] + '_marked'].append(dependent.copy())
                                    #print(class_choices[1] + '_marked', dependent)
                            else:
                                answer[class_choices[1] + '_marked'].append(dependent.copy())
                        elif ehto[-1].startswith('clemma.'): # compare lemma and word form
                            class_choices = ehto[-1][7:].split(';')
                            if lemma_comparison(dependent[1], dependent[2], kieli):
                                answer[class_choices[0] + '_marked'].append(dependent.copy())
                                #print(class_choices[0] + '_marked', dependent)
                            else:
                                answer[class_choices[1] + '_marked'].append(dependent.copy())
                                #print(class_choices[1] + '_marked', dependent)
                        else:
                            answer[ehto[-1]].append(dependent.copy())
                    elif dependent[7] == conj:
                        firstconj_i = find_head_index(dependent)
                        if firstconj_i >= 0:
                            firstconj = lause['analyysit'][head_i]
                            if firstconj in answer['dep_marked'] or head in answer['double_marked'] or head in answer['head_marked'] or head in answer['zero_marked']:
                                head_i = find_head_index(firstconj)
                                if head_i >= 0:
                                    dep_pos = dependent[3]
                                    dep_tag = dependent[5]
                                    dep_synt = conj
                                    head = lause['analyysit'][head_i]
                                    head_position = head[0]
                                    head_pos = head[3]
                                    head_tag = head[5]
                                    head = lause['analyysit'][head_i]
                                    lst = [dep_pos, dep_tag, dep_synt, head_pos, head_tag]
                                    new_conds = deepcopy(ehto[3:])
                                    new_conds[2] = '' # skip the conj
                                    to_append = dependent[:6] + [head_position] + dependent[7:] # keep original head length (of the one tagged nmod or something)
                                    # in order to calculate the dependency length correctly
                                    if all(compare(*pair) for pair in zip(lst, new_conds)):

                                        if '~' in ehto[-1]: # ----this block is never reached in these treebanks----
                                            plemma, ppos, psynt, instruction = ehto[-1].split('~') # psynt not applicable as it's a conj
                                            assert instruction.startswith('clemma.')
                                            class_choices = instruction[7:].split(';')
                                            
                                            do_lemma_comparison = True
                                            #print('potential conj-lemmacomp at sent {} ({}) in {}'.format(lause['sent_id'], dependent, kieli))

                                            for word in analyses: # Check if there is an article in the construction
                                                if word[6] != head[0]:
                                                    continue
                                                if (word[2] == plemma or not plemma) and (word[3] == ppos or not ppos):
                                                    do_lemma_comparison = False
                                                    break

                                            if do_lemma_comparison:
                                                
                                                if lemma_comparison(dependent[1], dependent[2], kieli):
                                                    answer[class_choices[0] + '_marked'].append(to_append)
                                                    #print(class_choices[0] + '_marked', dependent)
                                                else:
                                                    answer[class_choices[1] + '_marked'].append(to_append)
                                                    #print(class_choices[1] + '_marked', dependent)
                                            else:
                                                answer[class_choices[1] + '_marked'].append(to_append)
  
                                        elif ehto[-1].startswith('clemma.'): # compare lemma and word form
                                            class_choices = ehto[-1][7:].split(';')
                                            if lemma_comparison(dependent[1], dependent[2], kieli):
                                                answer[class_choices[0] + '_marked'].append(to_append)
                                                #print(class_choices[0] + '_marked', dependent)
                                            else:
                                                answer[class_choices[1] + '_marked'].append(to_append)
                                                #print(class_choices[1] + '_marked', dependent)
                                        else:
                                            answer[ehto[-1]].append(to_append)
                        
            else: # head_exist ("pyöränsä")
                for head in analyses:
                    head_pos = head[3]
                    head_tag = head[5]
                    lst = [head_pos, head_tag]
                    nmodposs = ehto[5].split(';') # nmod or nmod:poss, NOT: "on.nmod" nor "on.nmod:poss"
                    if all(compare(*pair) for pair in zip(lst, ehto[6:])):
                        if all([(word[6] != head[0] or word[7] not in nmodposs) for word in analyses]):
                            answer[ehto[-1]].append(head.copy())
        else:
            for word in analyses:
                part_index = int(word[0])
                part_lemma = word[2]
                part_pos = word[3]
                part_synt = word[7]
                dep_i = find_head_index(word)
                if dep_i < 0:
                    part_index = frozenset()
                    continue
                dependent = lause['analyysit'][dep_i]
                dep_pos = dependent[3]
                dep_tag = dependent[5]
                dep_synt = dependent[7]
                head_i = find_head_index(dependent)
                if head_i < 0:
                    part_index = frozenset()
                    continue
                head = lause['analyysit'][head_i]
                head_pos = head[3]
                head_tag = head[5]

                lst = [part_lemma, part_pos, part_synt, dep_pos, dep_tag, dep_synt, head_pos, head_tag]

                if all(compare(*pair) for pair in zip(lst, ehto)):
                    if ehto[-1].startswith('clemma.'): # compare lemma and word form
                        class_choices = ehto[-1][7:].split(';')
                        if lemma_comparison(dependent[1], dependent[2], kieli):
                            answer[class_choices[0] + '_marked'].append(dependent + [part_index])
                            #print(class_choices[0] + '_marked', dependent)
                        else:
                            answer[class_choices[1] + '_marked'].append(dependent + [part_index])
                            #print(class_choices[1] + '_marked', dependent)
                    else:
                        answer[ehto[-1]].append(dependent + [part_index])

                    #collect conjs for this dependent with an adposition
                    #won't work if the adposition is a from a different condition
                    for wd in reversed(analyses):
                        if wd[7] != conj: # "analyses" is sorted so that conjs are at the end
                            break
                        elif wd[6] != dependent[0]:
                            continue
                        firstconj = dependent
                        head_position = head[0]
                        to_append = wd[:6] + [head_position] + wd[7:] # keep original head length (of the one tagged nmod or something)

                        new_lst = [wd[3], wd[5], dep_synt, head_pos, head_tag]
                        new_conds = deepcopy(ehto[3:])

                        if all(compare(*pair) for pair in zip(new_lst, new_conds)):
                            if ehto[-1].startswith('clemma.'): # compare lemma and word form
                                class_choices = ehto[-1][7:].split(';')
                                if lemma_comparison(dependent[1], dependent[2], kieli):
                                    answer[class_choices[0] + '_marked'].append(to_append + [part_index])
                                    #print(class_choices[0] + '_marked', dependent)
                                else:
                                    answer[class_choices[1] + '_marked'].append(to_append + [part_index])
                                    #print(class_choices[1] + '_marked', dependent)
                            else:
                                answer[ehto[-1]].append(to_append + [part_index])
                part_index = frozenset()

    for marking in answer:
        if marking == 'head_exist':
            for n, head in enumerate(answer[marking]):
                answer[marking][n].append(0)
        else:
            for n, dep in enumerate(answer[marking]):

                #lst[m:n] will return from m+1th element up to nth.
                #since we want from this+1st to head-1st, or head+1st to this-1st,
                #1 will be subtracted from the LARGER number in order to get the correct indices.
                
                
                this_i = int(dep[0])
                head_i = int(dep[6])
                # only those in between will be checked
                if this_i > head_i:
                    this_i -= 1
                elif this_i < head_i:
                    head_i -= 1
                else:
                    raise AssertionError("Head and dependent have the same ordinal: {} from {}".format(kieli, lause['sent_id']))
                if not (this_i >= 0 and head_i >= 0):
                    raise AssertionError("Head or dependent is root: {} from {}".format(kieli, lause['sent_id']))

                start = min(this_i, head_i)
                stop = max(this_i, head_i)

                answer[marking][n].append(len([w for w in lause['analyysit'][start:stop] if w[3] == punct]))
                
    
    return defaultdict(lambda:[], remove_duplicates(answer))
    #return remove_duplicates(answer) #does not work

tabmodel = read_table('ud_tabmodel.txt')

all_languages = sorted(set(languages + hard_languages))

def mean(x): # Count mean dependency lenghths with this mean function. If no such marking exists, return NA.
    if len(x) == 0:
        return NA
    return sum(x)/len(x)


punctlog = sys.stdout

punctlog_set = set()

def ii0(x, material):
    global punctlog_set
    if type(x) == int:
        if x < 0:
            print('Warning: negative dependency length in {}.'.format(material))
        return x
    if material not in punctlog_set:
        print(material, file=punctlog)
        punctlog_set.add(material)
    return 0

def format_with_punctlog(thing, material):
    if material in punctlog_set:
        return thing + ' (puncts not subtracted yet)'
    else:
        return thing

def is_in_between(adp, w1, w2):
    #print(adp, w1, w2)
    if type(adp) != int:
        return 0
    assert adp != w1 and adp != w2
    n = min(int(w1), int(w2))
    x = max(int(w1), int(w2))
    return n < adp < x

def firstfirst(head_ord, dependent_ord):
    try:
        ho = int(head_ord)
        do = int(dependent_ord)
        if ho < do:
            return 'hd'
        elif do < ho:
            return 'dh'
        else:
            return NA
    except ValueError:
        return NA


allposs_filename = 'FullNPPossData.txt'

def save(name=allposs_filename):
    """Analyze all materials and save them to a tab-separated file."""
    allposs = open(name, 'wt', encoding='utf-8')
    print('poss_no', 'id', 'language', 'type', 'length1', 'length2', 'sent_id', 'dep_ord', 'head_ord', 'adp_ord', 'N#puncts', 'order', 'conj', sep='\t', file=allposs)#, 'sent_id', 'list', sep='\t', file=allposs)
####0: poss_no (from 0 to the number of possessive NPs in all treebanks)
####1: id (treebank ID, e.g. "afr", "fin_tdt")
####2: language (language name, eg. "English", "Old Church Slavonic")
####3: type (type of marking, one of the following: "dep_marked" / "double_marked" / "head_marked" / "zero_marked" / "head_only")
####4: length1 ([absolute value of dep_ord and head_ord] minus [N#puncts])
####5: length2 ([absolute value of dep_ord and head_ord] minus [N#puncts] minus [1 if adp_ord is between dep_ord and head_ord, 0 otherwise])
####6: sent_id (sentence ID in the treebank)
####7: dep_ord (ordinal of dependent (=possessor) in the sentence (NA if not present, in which case type is "head_only"))
####8: head_ord (ordinal of dependent (=possessee) in the sentence)
####9: adp_ord (ordinal of adposition/particle equivalent of English 'of' in the sentence; NA if not present)
####10: N#puncts (number of puncts (POS tag == "PUNCT") between dependent and head)
####11: first (which one is first in the sentence, the head or dependent; NA if not present)
####12: conj (is this a conj or not)
    n = 1
        #print('id\tlanguage\tdep_marked\tdouble_marked\thead_marked\tzero_marked\thead_exist\tdep_dl_mean\tdouble_dl_mean\thead_dl_mean\tzero_dl_mean\tmean_dl	size', file=file)
    for language in tabmodel[1:]:
        material = language[2]
        print('Working on {}'.format(language[2]), end=' ')
        starttime = monotonic()
        dep_distances = []
        double_distances = []
        head_distances = []
        zero_distances = []

        distances = []
        
        if language[2][3:] not in hard_languages and not is_done(language[1].replace(' ', '_')):
            counts = {'dep_marked':NA, 'double_marked':NA, 'head_marked':NA, 'zero_marked':NA, 'head_exist':NA}
            size = import_data(material, 'size')
        else:
            d, size = import_data(material, 'both')
            items = []
            ids = []
            for i in d:
                od = to_ordered_dict(i)
                items.append(possessiivi(od, language[2][3:]))
                ids.append(od['sent_id'])
            #items = [possessiivi(to_ordered_dict(i), language[2][3:]) for i in d]
            for itemindex, defdict in enumerate(items):
                for marking in defdict:
                    if marking == 'head_exist':
                        for found in defdict[marking]:
#'poss_no', 'id', 'language', 'type', 'length1', 'length2', 'sent_id', 'dep_ord', 'head_ord', 'adp_ord', 'N#puncts', 'first', 'conj'
                            print(n, language[0], language[1], 'head_only', NA, NA, ids[itemindex], NA, found[0], NA, NA, found[7]=='conj', sep='\t', file=allposs)
                            n += 1
                    else:
                        for found in defdict[marking]:
                            in_between = is_in_between(found[-2], found[0], found[6])
                            print(n, language[0], language[1], marking,
                                  abs(int(found[0])-int(found[6]))-found[-1], #length1
                                  abs(int(found[0])-int(found[6]))-found[-1]-in_between, #length2
                                  ids[itemindex], found[0], found[6],
                                  found[-2] if type(found[-2]) == int else NA, # adp_ord; NA if not applicable
                                  found[-1],
                                  firstfirst(head_ord=found[6], dependent_ord=found[0]),
                                  found[7]=='conj', sep='\t', file=allposs)
                            n += 1
                    
        print('took {} seconds.'.format(format(monotonic()-starttime, '.3f')))
    allposs.close()


def str2bea(s, language): # for the GUI
    """Returns a string representing the dependents of each possessive NP of a sentence s
(as a string representing the conllu item) and language (material)."""
    answer = []
    d = possessiivi(to_ordered_dict(s), language)

    for item in ['dep_marked', 'double_marked', 'head_marked', 'zero_marked', 'head_exist']:
        if d[item] == NA:
            answer.append('{}: {}'.format(item, d[item]))
        else:
            answer.append('{}: {} ({} kpl)'.format(item, d[item], len(d[item])))

    return '\n'.join(answer)

def cint(x):
    if x == NA:
        return NA
    return int(x)

def sort_numberset(x):
    if NA not in x:
        return sorted([int(i) for i in x])
    else:
        return sorted([int(i) for i in x if i != NA]) + ['NA']


if __name__ == '__main__':# and 'idlelib' in sys.modules:
    save() # save with duplicates and then overwrite


    # If you want to keep the file with duplicates, comment out the following
    # lines all the way until the comment "Duplicate overwriting end"
    # or alternatively change the variable "allposs_filename" to some other
    # string representing a valid filename
    print('File with duplicates saved, now replacing it with a file without them...')

    no_duplicates = defaultdict(lambda:set())

    taul = read_table(allposs_filename)

    new_title_row = taul[0].copy()
    new_title_row[9] += 's' # adp_ord -> adp_ords

    for item in taul[1:]:
        no_duplicates[tuple(item[1:5] + item[6:9] + item[10:])].add(item[9])

    n = 1

    with open(allposs_filename, 'wt', encoding='utf-8') as k:
        print('\t'.join(new_title_row), file=k)
        for jtem in no_duplicates:
            print(n, end='\t', file=k)
            print('\t'.join(jtem[:4]), end='\t', file=k)
            if jtem[5] == NA:
                length2 = NA
                adp_ords = [NA]
            else:
                dep_ord = int(jtem[5])
                head_ord = int(jtem[6])
                adp_ords = sort_numberset(no_duplicates[jtem])
                length2 = abs(dep_ord-head_ord)
                for ordinal in adp_ords: # subtract adpositions if in between
                    length2 -= is_in_between(ordinal, dep_ord, head_ord)
            try:
                print(length2-int(jtem[7]), end='\t', file=k)
            except ValueError: # head_only NA
                print(NA, end='\t', file=k)
            print('\t'.join(jtem[4:7]), end='\t', file=k)
            print(','.join([str(z) for z in adp_ords]), end='\t', file=k)
            print('\t'.join(jtem[7:]), file=k)

            n += 1

    print('Done')
    # Duplicate overwriting end
    
    if punctlog != sys.stdout:
        punctlog.close()
    elif 'idlelib' not in sys.modules:
        ##input('Done. Press Enter to exit.')
        sleep(5) # let the user see the print output for five seconds before the window closes (if it's going to close)

    
